<?php

namespace App\Livewire\Owner;

use App\Models\Paiement;
use App\Models\Product;
use App\Models\Sale;
use App\Models\SaleItem;
use App\Models\SupplyItem;
use App\Models\UsedNumber;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Jantinnerezo\LivewireAlert\LivewireAlert;
use Livewire\Attributes\On;
use Livewire\Component;
use Livewire\Attributes\Url;
use Livewire\WithPagination;

class SalesComponent extends Component
{
    use WithPagination, LivewireAlert;

    #[Url]
    public $search = '';
    public $isModalOpen = false;
    public $isModalFacturerOpen = false;
    public $newSale = null;
    public $saleItems = [];
    public $total_amount = 0;
    public $products = [];
    public $saleIdToDelete = null;
    public $saleIdToFacturer = null;
    public $saleToFacturer = null;
    public $productItemToAdd = null;
    public $quantityToAdd = 0;
    public $priceToAdd = 0;
    public $encashed_amount = 0;
    public $type_sale = 'paiement';
    public $client_name, $client_phone;
    public $mode_paiement_selected = null;
    public $modes_paiement = [[ 'name' => 'Cash', 'image' => 'cash.png' ],
    [ 'name' => 'Wave', 'image' => 'wave.png' ],[ 'name' => 'Orange Money', 'image' => 'orange-money.png' ],
    [ 'name' => 'Cheque', 'image' => 'cheque.png' ],[ 'name' => 'Virement', 'image' => 'virement.png' ],
    [ 'name' => 'Autre', 'image' => 'other.png' ] ];
     
    protected $rules = [
        'newSale.discount' => 'nullable|numeric|min:0',
        'newSale.tax' => 'nullable|numeric|min:0',
        'newSale.total' => 'nullable|numeric|min:0',
    ];

    private function calculateTotal($items): int
    {
        return array_sum(array_map(
            fn($item) => intval($item['price']) * intval($item['quantity']),
            $items
        ));
    }

    private function get_last_or_number(): int
    {
        return DB::transaction(function () {
            $lastNumber = UsedNumber::lockForUpdate()
                ->where('shop_id', Auth::user()->shop_id)
                ->orderByDesc('number')
                ->first();

            $nextNumber = $lastNumber ? $lastNumber->number + 1 : 10001;

            UsedNumber::create([
                'number' => $nextNumber,
                'shop_id' => Auth::user()->shop_id
            ]);

            return $nextNumber;
        });
    }

    public function mount()
    {
        $this->products = Product::all();
    }

    public function addItem()
    {
        if(!$this->productItemToAdd) {
            $this->alert('error', 'Veuillez sélectionner un produit');
            return;
        }

        $product = Product::find($this->productItemToAdd);
        $existingItemKey = array_search($this->productItemToAdd, array_column($this->saleItems, 'product_id'));
        
        if ($existingItemKey !== false) {
            $this->saleItems[$existingItemKey]['quantity']++;
        } else {
            $this->saleItems[] = [
                'product_name' => $product->name,
                'product_id' => $this->productItemToAdd,
                'quantity' => 1,
                'price' => $product->price,
            ];
        }

        $this->total_amount = $this->calculateTotal($this->saleItems);
        $this->productItemToAdd = null;
        $this->dispatch('clear-product-select');
    }

    public function removeItem($index)
    {
        unset($this->saleItems[$index]);
        $this->saleItems = array_values($this->saleItems);
        $this->total_amount = $this->calculateTotal($this->saleItems);
    }

    public function updatedSaleItems()
    {
        $this->total_amount = $this->calculateTotal($this->saleItems);
    }

    #[On('update-data')] 
    public function updateData($data)
    {
        if(isset($data['id'])) {
            $this->productItemToAdd = $data['id'];
        }
    }

    public function render()
    {
        $sales = Sale::withSum('paiements', 'amount')->orderByDesc('created_at')->paginate(30);
        return view('livewire.owner.sales-component', [
            'sales' => $sales,
        ]);
    }

    public function openModal($saleId = null)
    {
        $this->newSale = Sale::find($saleId);

        if (!$this->newSale) {
            $this->newSale = new Sale();
            $this->newSale->tax = 0;
            $this->newSale->discount = 0;
        } else {
            $items = SaleItem::with('product')
                ->where('sale_id', $this->newSale->id)
                ->get();

            $this->saleItems = $items->map(fn($item) => [
                'product_name' => $item->product->name,
                'product_id' => $item->product_id,
                'quantity' => $item->quantity,
                'price' => $item->price,
            ])->toArray();
        }

        $this->isModalOpen = true;
    }

    public function save()
    {
        if(count($this->saleItems) <= 0){
            $this->alert('error', 'Veuillez ajouter des produits à la vente');
            return;
        }
        $validated = $this->validate();
        $sale_to_save = new Sale();
        if(!empty($this->newSale->id)){
            $sale_to_save = Sale::find($this->newSale->id);
        }
        $sale_to_save->fill($validated);
        $total_amount = $this->calculateTotal($this->saleItems);
        $discount = $sale_to_save->discount ?? 0;
        $total_after_discount = $total_amount - $discount;
        $tax = $sale_to_save->tax ?? 0;
        $total_with_tax = $total_after_discount + ($tax > 0 ? ($total_after_discount * $tax/100) : 0);
        $sale_to_save->total = $total_amount;
        $sale_to_save->user_id = Auth::id();

        if(!$sale_to_save->invoice_number) {
            $sale_to_save->invoice_number = $this->get_last_or_number();
        }
        $sale_to_save->total_after_discount = $total_after_discount;
        $sale_to_save->total_with_tax = $total_with_tax;
        DB::beginTransaction();
        try {

            $sale_to_save->save();
            
            SaleItem::where('sale_id', $sale_to_save->id)->delete();
            
            $saleItems = array_map(fn($item) => [
                'sale_id' => $sale_to_save->id,
                'product_id' => $item['product_id'],
                'quantity' => $item['quantity'],
                'price' => $item['price']
            ], $this->saleItems);
            
            SaleItem::insert($saleItems);
            //update product stock
            foreach($this->saleItems as $item){
                $supplyItem = SupplyItem::where('product_id', $item['product_id'])->where('stock', '>', 0)->first();
                if($supplyItem){
                    $supplyItem->stock -= $item['quantity'];
                    $supplyItem->save();
                }
               /*  $product = Product::find($item['product_id']);
                $product->stock -= $item['quantity'];
                $product->save(); */
            }

            DB::commit();
            $this->isModalOpen = false;
            $this->reset();
            $this->alert('success', 'Vente enregistrée avec succès');
            
        } catch (\Exception $e) {
            DB::rollback();
            dd($e);
            $this->alert('error', 'Une erreur est survenue lors de l\'enregistrement');
        }
    }


    public function showModalFacturer($saleId)
    {
        $this->saleIdToFacturer = $saleId;
        $this->isModalFacturerOpen = true;
        $this->saleToFacturer = Sale::find($saleId);
        $this->encashed_amount = intval($this->saleToFacturer->total_with_tax);
    }

    public function delete($saleId)
    {
        $this->saleIdToDelete = $saleId;
        $this->confirm('Voulez-vous vraiment supprimer cette vente ?', [
            'onConfirmed' => 'confirmed',
        ]);
    }

    #[On('confirmed')]
    public function confirmDelete()
    {
        Sale::findOrFail($this->saleIdToDelete)->delete();
        $this->reset('saleIdToDelete');
        $this->alert('success', 'Vente supprimée avec succès');
    }

    public function saveFacture()
    {
        if($this->type_sale == 'paiement'){
            if(!$this->mode_paiement_selected){
                $this->alert('error', 'Veuillez sélectionner un mode de paiement');
                return;
            }
            if($this->encashed_amount <0){
                $this->alert('error', 'Le montant encaissé ne peut pas être négatif');
                return;
            }
            
            
        }elseif($this->type_sale == 'creance'){
            if(!$this->client_name){
                $this->alert('error', 'Veuillez entrer le nom du client');
                return;
            }
            if(!$this->client_phone){
                $this->alert('error', 'Veuillez entrer le numéro de téléphone du client');
                return;
            }
        }
        $this->saleToFacturer->type_sale = $this->type_sale;
        $this->saleToFacturer->client_name = $this->client_name;
        $this->saleToFacturer->client_phone = $this->client_phone;
        $this->saleToFacturer->save();
        if($this->type_sale == 'paiement'){
            $payment = new Paiement();
            $payment->sale_id = $this->saleToFacturer->id;
            $payment->amount = $this->encashed_amount;
            $payment->payment_mode = $this->mode_paiement_selected;
           
            $payment->save();
        }
        $this->alert('success', 'Vente enregistrée avec succès');
        $this->isModalFacturerOpen = false;
        $this->reset();
    }
}
