<?php

namespace App\Livewire\Owner;

use App\Models\Supply;
use App\Models\SupplyItem;
use App\Models\Supplier;
use App\Models\Product;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Jantinnerezo\LivewireAlert\LivewireAlert;
use Livewire\Attributes\On;
use Livewire\Component;
use Livewire\Attributes\Url;
use Livewire\WithPagination;

class SupplyComponent extends Component
{
    use WithPagination;
    use LivewireAlert;

    #[Url]
    public $search = '';
    public $isModalOpen = false;
    public $supplyItem = null;
    public $supplyItems = [];
    public $supplyIdToDelete = null;
    public $selectedSupplier = null;
    public $selectedProduct = null;
    public $productItemToAdd = null;
    public $quantityToAdd = 1;
    public $priceToAdd = 0;
    public $total_amount = 0;
    public $supplier_name = null, $order_date = null;
    public $new_supplier_name = null, $new_supplier_phone = null;
    public $step = 1;

    protected $rules = [
        'supplyItem.supplier_id' => 'required|exists:suppliers,id',
        'supplyItem.order_date' => 'required|date',
/*         'supplyItem.status' => 'required|in:pending,delivered,cancelled',
 */        'supplyItems.*.product_id' => 'required|exists:products,id',
        'supplyItems.*.quantity' => 'required|integer|min:1',
        'supplyItems.*.price' => 'required|numeric|min:0'
    ];

    public function mount()
    {
        /* $this->supplyItems = [
            [
                'product_name' => 'Nom Produit',
                'product_id' => '2',
                'quantity' => '120',
                'price' => '120000'
            ]
        ];
        $this->supplyItem = new Supply(); */
    }

    public function render()
    {
        return view('livewire.owner.supply-component', [
            'items' => Supply::where('order_number', 'like', '%'.$this->search.'%')
                ->with(['supplier', 'items.product', 'shop'])
                ->paginate(10),
            'suppliers' => Supplier::all(),
            'products' => Product::all()
        ]);
    }

    public function openModal($supplyId = null)
    {
        $this->supplyItem = Supply::find($supplyId);
        if (!$this->supplyItem) {
            $this->supplyItem = new Supply();
            $this->supplyItems = [];
        } else {
            $this->supplyItems = $this->supplyItem->items->toArray();
        }
        $this->isModalOpen = true;
    }
    public function updatedSupplyItem()
    {
        if($this->supplyItem->supplier_id){
            $this->supplier_name = Supplier::find($this->supplyItem->supplier_id)->name;
        }
        if($this->supplyItem->order_date){
            $this->order_date = date('d/m/Y', strtotime($this->supplyItem->order_date));
        }
    }

    public function addItem()
    {
        if (!$this->productItemToAdd) {
            $this->alert('error', 'Veuillez sélectionner un produit', [
                'timer' => null,
                'timerProgressBar' => true,
                'position' => 'top-end',
                'toast' => true,
            ]);
            return;
        }
        if ($this->quantityToAdd < 1) {
            $this->alert('error', 'Veuillez entrer une quantité');
            return;
        }
        if ($this->priceToAdd < 0) {
            $this->alert('error', 'Veuillez entrer un prix');
            return;
        }
        $productInChoosen = array_filter($this->supplyItems, function($item){
            return $item['product_id'] == $this->productItemToAdd;
        });
        if(count($productInChoosen) > 0){
            $this->alert('error', 'Le produit a déjà été ajouté, vous pouvez le modifier la quantité ou le prix');
            return;
        }
        $this->supplyItems[] = [
            'product_name' => Product::find($this->productItemToAdd)->name,
            'product_id' => $this->productItemToAdd,
            'quantity' => $this->quantityToAdd,
            'price' => $this->priceToAdd
        ];

        $this->productItemToAdd = null;
        $this->quantityToAdd = 1;
        $this->priceToAdd = 0;
        $this->dispatch('clear-product-select');
        $this->calculateTotal();
    }
    public function updatedSupplyItems()
    {
        //dd($this->supplyItems);
        $this->calculateTotal();
    }
    public function nextStep()
    {
        if($this->step == 1){
            if($this->supplyItem->supplier_id == null){
                $this->alert('error', 'Veuillez sélectionner un fournisseur');
                return;
            }
            if($this->supplyItem->supplier_id == 0){
                $this->validate([
                    'new_supplier_name' => 'required|string|max:255',
                    'new_supplier_phone' => 'required|string|max:255',
                ]);
                $supplier = Supplier::create([
                    'name' => $this->new_supplier_name,
                    'phone' => $this->new_supplier_phone,
                    'shop_id' => Auth::user()->shop_id
                ]);
                $this->supplyItem->supplier_id = $supplier->id;
            }
            $this->validate([
                'supplyItem.supplier_id' => 'required|exists:suppliers,id',
                'supplyItem.order_date' => 'required|date'
            ]);
            $this->step = 2;
        }else{
            if($this->step == 2){
                $this->step = 3;
            }
        }
        
        
    }
    public function previousStep()
    {
        if($this->step > 1){
            $this->step = $this->step - 1;
        }
    }
    
    public function removeItem($index)
    {
        unset($this->supplyItems[$index]);
        $this->supplyItems = array_values($this->supplyItems);
        
    }

    public function calculateTotal()
    {
        $total = 0;
        foreach ($this->supplyItems as $item) {
            $total += intval($item['quantity']) * floatval($item['price']);
        }
        $this->total_amount = $total;
    }

    public function save()
    {
        if($this->supplyItem->supplier_id == null){
            $this->alert('error', 'Veuillez sélectionner un fournisseur');
            return;
        }
        if($this->supplyItem->order_date == null){
            $this->alert('error', 'Veuillez sélectionner la date de commande');
            return;
        }
        /* if($this->supplyItem->status == null){
            $this->alert('error', 'Veuillez sélectionner le statut de la commande');
            return;
        } */
       if(count($this->supplyItems) == 0){
            $this->alert('error', 'Veuillez ajouter des produits à la commande');
            return;
       }
        $data = $this->validate();
        DB::beginTransaction();
        try {
            if (!$this->supplyItem->order_number) {
                $this->supplyItem->order_number = 'CMD-'.str_pad(Supply::count() + 1, 6, '0', STR_PAD_LEFT);
            }
            $this->supplyItem->total_amount = $this->total_amount;
            $this->supplyItem->shop_id = Auth::user()->shop_id;
            $this->supplyItem->save();
           
            foreach ($this->supplyItems as $item) {
                SupplyItem::create([
                    'supply_id' => $this->supplyItem->id,
                    'product_id' => $item['product_id'],
                    'quantity' => $item['quantity'],
                    'stock' => $item['quantity'],
                    'price' => $item['price']
                ]);
                $product = Product::find($item['product_id']);
                $product->stock += $item['quantity'];
                $product->save();
            }
            
            
            DB::commit();
            $this->isModalOpen = false;
            $this->reset();
            $this->alert('success', 'Commande enregistrée avec succès');
        } catch (\Exception $e) {
            DB::rollback();

            $this->alert('error', 'Une erreur est survenue lors de l\'enregistrement');
            return;
        }
        
       
    }
    #[On('update-data')]
    public function updateData($data)
    {
        $this->productItemToAdd = $data['id'];
    }


    public function delete($supplyId)
    {
        $this->supplyIdToDelete = $supplyId;
        $this->confirm('Voulez-vous vraiment supprimer cette commande ?', [
            'onConfirmed' => 'confirmed',
        ]);
        
    }

    #[On('confirmed')]
    public function confirmDelete()
    {
        DB::beginTransaction();
        try {
            // Récupérer les éléments de la commande avant suppression
            $supplyItems = Supply::find($this->supplyIdToDelete)->items;
            
            //dd($supplyItems);
            // Mettre à jour le stock pour chaque produit
            foreach ($supplyItems as $item) {
                $product = Product::find($item->product_id);
                $product->stock -= $item->quantity; // Déduire la quantité du stock
                $product->save();
            }
            $supply = Supply::find($this->supplyIdToDelete);
            $supply->items()->delete();
            $supply->delete();
            DB::commit();
            $this->reset('supplyIdToDelete');
            $this->alert('success', 'Commande supprimée avec succès');
        } catch (\Exception $e) {
            DB::rollback();
            dd($e);
            $this->alert('error', 'Une erreur est survenue lors de la suppression');
        }
    }
}
